#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
#
# semi-automatically prepare the upstreaming of the bpf files into a kernel
# tree branch.
#

from pathlib import Path
from dataclasses import dataclass
from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.theme import Theme
import click
import difflib
import filecmp
import git
import os
import rich
import shutil
import sys


DEFAULT_KERNEL_PATH = "../hid"


@dataclass
class Repos:
    kernel: git.Repo
    udev_hid_bpf: git.Repo

    @property
    def kernel_bpf_dir(self):
        return Path(self.kernel.working_tree_dir) / "drivers" / "hid" / "bpf" / "progs"


pass_repo = click.make_pass_decorator(Repos)
script_dir = Path(os.path.dirname(os.path.realpath(__file__)))


@click.group()
@click.option(
    "--kernel",
    type=click.Path(exists=True),
    default=Path(DEFAULT_KERNEL_PATH),
    help="Path to the hid.git kernel tree",
)
@click.option(
    "--udev-hid-bpf",
    type=click.Path(exists=True),
    default=script_dir.parent,
    help="Path to the udev-hid-bpf tree",
)
@click.pass_context
def cli(ctx, kernel, udev_hid_bpf):
    git_udev_hid_bpf = git.Repo(udev_hid_bpf)
    assert not git_udev_hid_bpf.bare
    git_kernel = git.Repo(kernel)
    assert not git_kernel.bare
    ctx.obj = Repos(git_kernel, git_udev_hid_bpf)


@cli.command()
@pass_repo
def to_kernel_tree(repos):
    """prepare commits on a kernel tree

    look for files in src/bpf/testing, fetch each file its git log
    copy those files to the kernel tree and commit the files with
    the content of the commit log.

    """
    if repos.udev_hid_bpf.head.ref.name not in ["main"]:
        click.confirm(
            f"current head ({repos.udev_hid_bpf.head.ref}) is not on 'main', are you sure?",
            abort=True,
        )

    click.confirm(
        f"currently on {repos.kernel.head.ref}, is that OK?",
        abort=True,
    )

    # gather commits creating/touching new files that need to be upstreamed
    tree = repos.udev_hid_bpf.head.commit.tree
    blobs = [
        b for b in (tree / "src/bpf/testing").traverse() if b.name != "meson.build"
    ]
    blobs.extend(
        [
            b
            for b in (tree / "src/bpf").traverse()
            if b.name.endswith(".h")
            and b.name not in ["attach.h", "vmlinux.h", "uhid-bpf-test-wrappers.h"]
        ]
    )

    history = {
        blob: repos.udev_hid_bpf.git.log("--pretty=%H", "--follow", blob.path).split(
            "\n"
        )
        for blob in blobs
    }

    # get committer information
    with repos.kernel.config_reader() as git_config:
        email = git_config.get_value("user", "email")
        user = git_config.get_value("user", "name")

    # copy the files, strip the '00xx-' prefix, and commit
    for blob, hist in history.items():
        source = Path(blob.abspath)
        dest = repos.kernel_bpf_dir / blob.name.lstrip("-0.123456789")

        # ignore if the file changed
        if Path(dest).exists() and filecmp.cmp(blob.abspath, dest, shallow=False):
            continue

        changes = []
        if Path(dest).exists():
            with open(dest) as f:
                k = f.readlines()

            for sha in hist:
                commit = repos.udev_hid_bpf.commit(sha)
                try:
                    obj = commit.tree / blob.path
                except KeyError:
                    # the file doesn't exist in this sha, but will be in the future
                    continue

                u = obj.data_stream.read().decode("utf-8").splitlines(keepends=True)

                diff = list(difflib.unified_diff(u, k))
                if not diff:
                    break

                changes.append(sha)

            hist = changes

            # not a 00XX- file, ask for confirmation
            if dest.name == blob.name:
                if not confirm_filediff(
                    blob.abspath,
                    str(dest),
                    f"changes detected in {blob.path}, do you want to backport them?",
                ):
                    continue

        message = "\n".join([repos.udev_hid_bpf.commit(hash).message for hash in hist])
        message += f"Signed-off-by: {user} <{email}>"

        rich.print(f"[green]copy[/green] {source} [green]to[/green] {dest}")
        shutil.copy(source, dest)
        repos.kernel.index.add([dest])
        repos.kernel.index.commit(message)


class DiffHighlighter(RegexHighlighter):
    """Apply style to anything that looks like a diff."""

    base_style = "diff."
    highlights = [
        r"(?P<deletion>^\-.*)",
        r"(?P<addition>^\+.*)",
    ]


def confirm_filediff(src: str, dst: str, message: str) -> bool:
    with open(src) as src_f, open(dst) as dst_f:
        s = src_f.readlines()
        d = dst_f.readlines()

    theme = Theme(
        {
            "diff.addition": "green",
            "diff.deletion": "red",
        }
    )
    console = Console(highlighter=DiffHighlighter(), theme=theme)
    for line in difflib.unified_diff(s, d, fromfile=src, tofile=dst):
        console.print(line.rstrip("\n"))
    rich.print("---")
    return click.confirm(
        message,
    )


@cli.command()
@pass_repo
def from_kernel_tree(repos):
    """backport commits from the kernel tree

    look for files in drivers/hid/bpf/progs in the kernel tree, look for
    any differences with src/bpf/stable, and import the changes here.

    """
    click.confirm(
        f"currently on {repos.kernel.head.ref}, is that OK?",
        abort=True,
    )

    # gather commits creating/touching new files that need to be upstreamed
    stable_tree = repos.udev_hid_bpf.head.commit.tree
    stable_blobs = [
        b
        for b in (stable_tree / "src/bpf/stable").traverse()
        if b.name != "meson.build"
    ]
    testing_blobs = [
        b
        for b in (stable_tree / "src/bpf/testing").traverse()
        if b.name != "meson.build"
    ]
    common_headers_blobs = [
        b for b in (stable_tree / "src/bpf").traverse() if b.name.endswith(".h")
    ]

    kernel_tree = repos.kernel.head.commit.tree
    kernel_blobs = [
        b
        for b in (kernel_tree / "drivers/hid/bpf/progs").traverse()
        if b.name not in ["Makefile", "README"]
    ]

    # get committer information
    with repos.udev_hid_bpf.config_reader() as git_config:
        email = git_config.get_value("user", "email")
        user = git_config.get_value("user", "name")

    # iterate over all matching files in the kernel tree
    for blob in kernel_blobs:
        confirm = False

        # gather matching stable and testing files
        stable_matches = [b.abspath for b in stable_blobs if b.name.endswith(blob.name)]
        stable_matches.sort(reverse=True)
        testing_matches = [
            b.abspath for b in testing_blobs if b.name.endswith(blob.name)
        ]
        testing_matches.sort(reverse=True)

        if not stable_matches and not testing_matches and blob.name.endswith(".h"):
            m = [b.abspath for b in common_headers_blobs if b.name == blob.name]
            if m:
                stable_matches = m
                confirm = True

        changes = not stable_matches or any(
            not filecmp.cmp(blob.abspath, m, shallow=False) for m in stable_matches
        )

        if testing_matches:
            if not filecmp.cmp(blob.abspath, testing_matches[0], shallow=False):
                changes = confirm_filediff(
                    testing_matches[0],
                    blob.abspath,
                    f"changes detected in {blob.path}, do you still want to backport them?",
                )

        if not changes:
            continue

        dest = None
        message = None
        idx = 10

        rich.print(f"[blue]gathering history of[/blue] {blob.path}")
        history = repos.kernel.git.log("--pretty=%H", blob.path).split("\n")

        if not stable_matches:
            rich.print(f"[red]new file:[/red] {blob.name}")
        else:
            if confirm:
                if not confirm_filediff(
                    stable_matches[0],
                    blob.abspath,
                    f"changes detected in {blob.path}, do you want to backport them?",
                ):
                    # abort backport of current file, and go to the next
                    continue

                dest = stable_matches[0]
            else:
                prev_file = Path(stable_matches[0]).name
                idx = int(prev_file.split("-")[0]) + 10

            with open(stable_matches[0]) as f:
                u = f.readlines()

            new_shas = []

            for sha in history:
                commit = repos.kernel.commit(sha)
                obj = commit.tree / blob.path

                k = obj.data_stream.read().decode("utf-8").splitlines(keepends=True)

                diff = list(difflib.unified_diff(u, k))
                if not diff:
                    break

                new_shas.append(sha)

            if not new_shas:
                commit = repos.kernel.commit(history[0])
                obj = commit.tree / blob.path

                committed_content = (
                    obj.data_stream.read().decode("utf-8").splitlines(keepends=True)
                )
                with open(blob.abspath) as f:
                    current = f.readlines()

                sys.stdout.writelines(
                    difflib.unified_diff(
                        committed_content,
                        current,
                        fromfile=blob.path,
                        tofile="uncommitted content",
                    )
                )
                rich.print("---")

                if not click.confirm(
                    f"Uncommitted changes in {blob.path}, do you want to backport them?",
                ):
                    # abort backport of current file, and go to the next
                    continue

            history = new_shas

        if not dest:
            dest = (
                Path(repos.udev_hid_bpf.working_tree_dir)
                / "src"
                / "bpf"
                / "stable"
                / f"{idx:04}-{blob.name}"
            )

        message = "\n".join(
            [repos.kernel.commit(hash).message.split("---")[0] for hash in history]
        )
        if not message:
            message = "Uncommitted changes\n\n"
        for hash in history:
            message += (
                f"Upstream commit: https://git.kernel.org/hid/hid/c/{hash[:12]}\n"
            )
        message += f"Signed-off-by: {user} <{email}>"

        rich.print(f"[green]copy[/green] {blob.path} [green]into[/green] {dest}")
        shutil.copy(blob.abspath, dest)
        repos.udev_hid_bpf.index.add([dest])

        tracing_capable = False

        if testing_matches:
            meson_build = (
                Path(repos.udev_hid_bpf.working_tree_dir)
                / "src/bpf/testing/meson.build"
            )
            with open(meson_build) as f:
                meson = f.readlines()
            testing = [Path(file).name for file in testing_matches]
            with open(f"{meson_build}.new", "w") as f:
                in_tracing = False
                for line in meson:
                    found = False
                    if line.strip().startswith("tracing_sources = ["):
                        in_tracing = True
                    if in_tracing and line.strip() == "]":
                        in_tracing = False
                    for file in testing:
                        if file in line:
                            found = True
                            tracing_capable = in_tracing
                            break
                    if not found:
                        f.write(line)
            shutil.move(f"{meson_build}.new", meson_build)
            repos.udev_hid_bpf.index.add([meson_build])

            repos.udev_hid_bpf.index.remove(testing_matches, working_tree=True)

        # manually insert the new file in meson.build:
        # we iterate over meson.build, and when we are
        # between "sources = [" and "]", we gather the
        # list of files, add our current file, sort it,
        # and then re-inject everything
        meson_build = (
            Path(repos.udev_hid_bpf.working_tree_dir) / "src/bpf/stable/meson.build"
        )
        with open(meson_build) as f:
            meson = f.readlines()
        with open(f"{meson_build}.new", "w") as f:
            stable_files = [dest.name]
            found = False
            match = "tracing_sources = [" if tracing_capable else "sources = ["
            for line in meson:
                if line.strip().startswith(match):
                    f.write(line)
                    found = True
                elif found:
                    if line.strip() == "]":
                        stable_files.sort()
                        for n in stable_files:
                            f.write(f"    '{n}',\n")
                        f.write(line)
                        found = False
                    else:
                        stable_files.append(line.strip().strip("',"))
                else:
                    f.write(line)

        shutil.move(f"{meson_build}.new", meson_build)
        repos.udev_hid_bpf.index.add([meson_build])
        repos.udev_hid_bpf.index.commit(message)


if __name__ == "__main__":
    cli()



import git
import pytest
import re
import sys
import uuid
from dataclasses import dataclass
from pathlib import Path

try:
    from click.testing import CliRunner
except ModuleNotFoundError:
    pytest.skip("click is required for these tests", allow_module_level=True)

import sync_with_kernel_tree

from importlib.metadata import version as get_version
from packaging import version

click_pkg_version = get_version("click")
if version.parse(click_pkg_version) >= version.parse("8.2.0"):
    CLI_RUNNER_KWARGS = {}
else:
    CLI_RUNNER_KWARGS = {"mix_stderr": False}


@dataclass
class VirtualGit:
    kernel: git.Repo
    udev_hid_bpf: git.Repo

    @property
    def kernel_dir(self):
        """The kernel directory."""
        return Path(self.kernel.working_dir)

    @property
    def kernel_hid_bpf_dir(self):
        """The kernel hid-bpf directory."""
        return self.kernel_dir / "drivers" / "hid" / "bpf" / "progs"

    @property
    def udev_hid_bpf_dir(self):
        """The udev_hid_bpf directory."""
        return Path(self.udev_hid_bpf.working_dir)

    @property
    def udev_hid_bpf_bpf_dir(self):
        """The udev_hid_bpf src/bpf directory."""
        return self.udev_hid_bpf_dir / "src" / "bpf"

    def reset(self):
        self.kernel.head.reset(
            self._kernel_initial_commit, index=True, working_tree=True
        )
        for path in self.kernel.untracked_files:
            (self.kernel_dir / path).unlink()
        self.udev_hid_bpf.head.reset(
            self._udev_hid_bpf_initial_commit, index=True, working_tree=True
        )
        for path in self.udev_hid_bpf.untracked_files:
            (self.udev_hid_bpf_dir / path).unlink()


def meson_build_template(tracing_sources=[], sources=[]) -> str:
    tr_srcs = "\n".join([f"    '{e}'," for e in tracing_sources])
    if tracing_sources:
        tr_srcs += "\n"
    srcs = "\n".join([f"    '{e}'," for e in sources])
    if sources:
        srcs += "\n"
    return f"""# garbage before sources and tracing_sources

tracing_sources = [
{tr_srcs}]

# some more comments

sources = [
{srcs}]

# rest needs to be there too
"""


@pytest.fixture(scope="session")
def virtual_git(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("git_test_data")
    kernel = git.Repo.init(tmp_path / "kernel", b="main")
    uhb = git.Repo.init(tmp_path / "uhb", b="main")
    vg = VirtualGit(kernel, uhb)

    for repo in [kernel, uhb]:
        with repo.config_writer() as git_config:
            repo_name = Path(repo.working_tree_dir).name
            git_config.set_value("user", "email", f"pytest-{repo_name}@example.com")
            git_config.set_value("user", "name", f"Pytest {repo_name}")

    # initialize udev-hid-bpf fake repo with meson.build and a few files
    files = ["0010-first.bpf.c", "0010-second.bpf.c", "0010-third.bpf.c"]
    for dir in ["stable", "testing"]:
        path = vg.udev_hid_bpf_bpf_dir / dir
        path.mkdir(parents=True)
        meson_build = path / "meson.build"
        txt = (
            meson_build_template(files[:2], files[2:])
            if dir == "stable"
            else meson_build_template()
        )
        meson_build.write_text(txt)
        uhb.index.add([meson_build])
    uhb.index.commit("Initial commit")

    for filename in files:
        new_bpf_file = vg.udev_hid_bpf_bpf_dir / "stable" / filename
        new_bpf_file.write_text(f"random text for {filename}")
        vg.udev_hid_bpf.index.add(new_bpf_file)
        uhb_commit_message = f"initial import of {filename}\n\nSob: me"
        vg.udev_hid_bpf.index.commit(uhb_commit_message)

    vg._udev_hid_bpf_initial_commit = vg.udev_hid_bpf.head.commit

    # initialize kernel fake repo with a few knwon files
    vg.kernel_hid_bpf_dir.mkdir(parents=True)
    for filename in ["Makefile", "README"]:
        file = vg.kernel_hid_bpf_dir / filename
        file.touch()
        kernel.index.add(file)
    kernel.index.commit("initial commit")

    for filename in ["first.bpf.c", "second.bpf.c", "third.bpf.c"]:
        new_bpf_file = vg.kernel_hid_bpf_dir / filename
        new_bpf_file.write_text(f"random text for 0010-{filename}")
        vg.kernel.index.add(new_bpf_file)
        uhb_commit_message = f"initial import of {filename}\n\nSob: me"
        vg.kernel.index.commit(uhb_commit_message)

    vg._kernel_initial_commit = vg.kernel.head.commit

    return vg


def run_cli(virtual_git, command, args):
    _args = [
        "--kernel",
        virtual_git.kernel_dir,
        "--udev-hid-bpf",
        virtual_git.udev_hid_bpf_dir,
        command,
    ]
    _args.extend(args)
    runner = CliRunner(**CLI_RUNNER_KWARGS)
    return runner.invoke(
        sync_with_kernel_tree.cli,
        _args,
        input="y\n",
    )


@pytest.mark.parametrize("command", ["to-kernel-tree", "from-kernel-tree"])
def test_setup(virtual_git, command):
    """calling the script on already sync-ed trees is a no-op"""
    virtual_git.reset()
    kernel_commit = virtual_git.kernel.head.commit
    udev_hid_bpf_commit = virtual_git.udev_hid_bpf.head.commit
    result = run_cli(virtual_git, command, [])
    print(result.stdout)
    print(result.stderr, file=sys.stderr)
    assert result.exit_code == 0
    assert kernel_commit == virtual_git.kernel.head.commit
    assert udev_hid_bpf_commit == virtual_git.udev_hid_bpf.head.commit


def test_to_kernel_new_file(virtual_git):
    """Add a new testing file and forward it to the kernel"""
    virtual_git.reset()
    kernel_commit = virtual_git.kernel.head.commit
    new_bpf_file = (
        virtual_git.udev_hid_bpf_bpf_dir / "testing" / "0010-Test__pytest.bpf.c"
    )
    text = str(uuid.uuid4())
    new_bpf_file.write_text(text)
    virtual_git.udev_hid_bpf.index.add(new_bpf_file)
    uhb_commit_message = (
        "initial import of 0010-Test__pytest.bpf.c\n\nSob: pytest marker\n"
    )
    virtual_git.udev_hid_bpf.index.commit(uhb_commit_message)
    udev_hid_bpf_commit = virtual_git.udev_hid_bpf.head.commit

    result = run_cli(virtual_git, "to-kernel-tree", [])

    print(result.stdout)
    print(result.stderr, file=sys.stderr)
    assert result.exit_code == 0

    # udev-hid-bpf should be untouched
    assert udev_hid_bpf_commit == virtual_git.udev_hid_bpf.head.commit

    # kernel has seen changes
    assert kernel_commit != virtual_git.kernel.head.commit
    assert uhb_commit_message in virtual_git.kernel.head.commit.message
    new_kernel_file = virtual_git.kernel_hid_bpf_dir / "Test__pytest.bpf.c"
    assert new_kernel_file.exists()
    assert new_kernel_file.read_text() == text


def test_to_kernel_update_file(virtual_git):
    """Update an existing stable file in testing, and forward it to the kernel"""
    virtual_git.reset()
    kernel_commit = virtual_git.kernel.head.commit
    bpf_file = virtual_git.udev_hid_bpf_bpf_dir / "testing" / "0020-first.bpf.c"
    text = str(uuid.uuid4())
    bpf_file.write_text(text)
    virtual_git.udev_hid_bpf.index.add(bpf_file)
    uhb_commit_message = "new version of 0020-first.bpf.c\n\nSob: pytest marker\n"
    virtual_git.udev_hid_bpf.index.commit(uhb_commit_message)
    udev_hid_bpf_commit = virtual_git.udev_hid_bpf.head.commit

    result = run_cli(virtual_git, "to-kernel-tree", [])

    print(result.stdout)
    print(result.stderr, file=sys.stderr)
    assert result.exit_code == 0

    # udev-hid-bpf should be untouched
    assert udev_hid_bpf_commit == virtual_git.udev_hid_bpf.head.commit

    # kernel has seen changes
    assert kernel_commit != virtual_git.kernel.head.commit
    assert uhb_commit_message in virtual_git.kernel.head.commit.message
    sob_indexes = [
        x.start() for x in re.finditer("Sob:", virtual_git.kernel.head.commit.message)
    ]
    assert len(sob_indexes) == 1
    kernel_file = virtual_git.kernel_hid_bpf_dir / "first.bpf.c"
    assert kernel_file.exists()
    assert kernel_file.read_text() == text


def test_from_kernel_new_file(virtual_git):
    """backport a file already in testing and now upstream into stable"""
    virtual_git.reset()

    # Add a pre-existing testfile
    testing_bpf_file = (
        virtual_git.udev_hid_bpf_bpf_dir / "testing" / "0010-Test__pytest.bpf.c"
    )
    text = str(uuid.uuid4())
    testing_bpf_file.write_text(text)
    virtual_git.udev_hid_bpf.index.add(testing_bpf_file)
    meson_build = virtual_git.udev_hid_bpf_bpf_dir / "testing" / "meson.build"
    meson_build.write_text(meson_build_template(["0010-Test__pytest.bpf.c"]))
    virtual_git.udev_hid_bpf.index.add([meson_build])
    uhb_commit_message = (
        "initial import of 0010-Test__pytest.bpf.c\n\nSob: pytest marker\n"
    )
    virtual_git.udev_hid_bpf.index.commit(uhb_commit_message)
    udev_hid_bpf_commit = virtual_git.udev_hid_bpf.head.commit

    # same file in the kernel
    new_bpf_file = virtual_git.kernel_hid_bpf_dir / "Test__pytest.bpf.c"
    new_bpf_file.write_text(text)
    virtual_git.kernel.index.add(new_bpf_file)
    uhb_commit_message += "Signed-off-by: pytest marker\n"
    virtual_git.kernel.index.commit(uhb_commit_message)
    kernel_commit = virtual_git.kernel.head.commit

    result = run_cli(virtual_git, "from-kernel-tree", [])

    print(result.stdout)
    print(result.stderr, file=sys.stderr)
    assert result.exit_code == 0

    # kernel should be untouched
    assert kernel_commit == virtual_git.kernel.head.commit

    # udev-hid-bpf has seen changes
    assert udev_hid_bpf_commit != virtual_git.udev_hid_bpf.head.commit
    assert uhb_commit_message in virtual_git.udev_hid_bpf.head.commit.message
    new_file = virtual_git.udev_hid_bpf_bpf_dir / "stable" / "0010-Test__pytest.bpf.c"
    assert new_file.exists()
    assert new_file.read_text() == text
    assert not testing_bpf_file.exists()

    # meson.build in testing is stripped from the old file
    assert meson_build_template() == meson_build.read_text()

    # meson.build in stable is enhanced with the new file
    meson_stable_build = virtual_git.udev_hid_bpf_bpf_dir / "stable" / "meson.build"
    assert (
        meson_build_template(
            ["0010-Test__pytest.bpf.c", "0010-first.bpf.c", "0010-second.bpf.c"],
            ["0010-third.bpf.c"],
        )
        == meson_stable_build.read_text()
    )


def test_from_kernel_update_file(virtual_git):
    """backport a file already in testing and stable, and push a new version"""
    virtual_git.reset()

    # Add a pre-existing testfile (the number is irrelevant)
    testing_bpf_file = (
        virtual_git.udev_hid_bpf_bpf_dir / "testing" / "0030-second.bpf.c"
    )
    text = str(uuid.uuid4())
    testing_bpf_file.write_text(text)
    virtual_git.udev_hid_bpf.index.add(testing_bpf_file)
    meson_build = virtual_git.udev_hid_bpf_bpf_dir / "testing" / "meson.build"
    meson_build.write_text(meson_build_template([], ["0030-second.bpf.c"]))
    virtual_git.udev_hid_bpf.index.add([meson_build])
    uhb_commit_message = (
        "new version of 0030-Test__pytest.bpf.c\n\nSob: pytest marker\n"
    )
    virtual_git.udev_hid_bpf.index.commit(uhb_commit_message)
    udev_hid_bpf_commit = virtual_git.udev_hid_bpf.head.commit

    # same file in the kernel
    new_bpf_file = virtual_git.kernel_hid_bpf_dir / "second.bpf.c"
    new_bpf_file.write_text(text)
    virtual_git.kernel.index.add(new_bpf_file)
    uhb_commit_message += "Signed-off-by: pytest marker\n"
    virtual_git.kernel.index.commit(uhb_commit_message)
    kernel_commit = virtual_git.kernel.head.commit

    result = run_cli(virtual_git, "from-kernel-tree", [])

    print(result.stdout)
    print(result.stderr, file=sys.stderr)
    assert result.exit_code == 0

    # kernel should be untouched
    assert kernel_commit == virtual_git.kernel.head.commit

    # udev-hid-bpf has seen changes
    assert udev_hid_bpf_commit != virtual_git.udev_hid_bpf.head.commit
    assert uhb_commit_message in virtual_git.udev_hid_bpf.head.commit.message
    new_file = virtual_git.udev_hid_bpf_bpf_dir / "stable" / "0020-second.bpf.c"
    assert new_file.exists()
    assert new_file.read_text() == text
    assert not testing_bpf_file.exists()
    old_file = virtual_git.udev_hid_bpf_bpf_dir / "stable" / "0010-second.bpf.c"
    assert old_file.exists()

    # ensure the commit history has only the new changes, not the old ones
    sob_indexes = [
        x.start()
        for x in re.finditer("Sob:", virtual_git.udev_hid_bpf.head.commit.message)
    ]
    assert len(sob_indexes) == 1

    # meson.build in testing is stripped from the old file
    assert meson_build_template() == meson_build.read_text()

    # meson.build in stable is enhanced with the new file
    meson_stable_build = virtual_git.udev_hid_bpf_bpf_dir / "stable" / "meson.build"
    assert (
        meson_build_template(
            ["0010-first.bpf.c", "0010-second.bpf.c"],
            ["0010-third.bpf.c", "0020-second.bpf.c"],
        )
        == meson_stable_build.read_text()
    )



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2024 Red Hat

from ctypes import (
    c_int,
)
from typing import Optional, Tuple, Type
from dataclasses import dataclass
from pathlib import Path
from enum import IntEnum

import base64
import binascii
import logging
import ctypes
import os
import json
import pytest
import random
import collections
import dataclasses
import errno
import subprocess

from .btf import Btf

logger = logging.getLogger(__name__)
random.seed()


_btf_cache: dict[str, Btf] = {}


def _load_btf(source: str) -> Btf:
    """Load and cache the BTF data for a BPF program source name.

    Raises ``FileNotFoundError`` if the ``.so.p`` directory does not exist.
    """
    if source not in _btf_cache:
        ld_path = os.environ.get("LD_LIBRARY_PATH")
        assert ld_path is not None
        sofile_dir = Path(ld_path) / f"libtest-{source}.so.p"
        if not sofile_dir.exists():
            raise FileNotFoundError(
                f"Unable to locate {sofile_dir}, assuming this BPF wasn't built"
            )
        _btf_cache[source] = Btf.load(list(sofile_dir.iterdir()))
    return _btf_cache[source]


# to be automatically field by BTF thanks to libbpf
class HidDevice(ctypes.Structure):
    cname = "hid_device"


# see struct hid_probe_args
class HidProbeArgs(ctypes.Structure):
    cname = "hid_bpf_probe_args"

    @property
    def rdesc_bytes(self) -> bytes:
        """
        Get the report descriptor bytes.

        Returns:
            The report descriptor as bytes (up to rdesc_size)
        """
        return bytes(self.rdesc[: self.rdesc_size])

    @rdesc_bytes.setter
    def rdesc_bytes(self, rdesc_data: bytes | bytearray):
        """
        Set the report descriptor data and size.

        Args:
            rdesc_data: The report descriptor bytes to set

        Example:
            probe_args = HidProbeArgs()
            probe_args.rdesc_bytes = bytearray([0x05, 0x01, 0x09, 0x02, ...])
        """
        for i, byte in enumerate(rdesc_data[: len(self.rdesc)]):
            self.rdesc[i] = byte
        self.rdesc_size = min(len(rdesc_data), len(self.rdesc))


# see struct hid_bpf_ctx
class HidBpfCtx(ctypes.Structure):
    cname = "hid_bpf_ctx"


class BpfTimer(ctypes.Structure):
    cname = "bpf_timer"


class BpfWq(ctypes.Structure):
    cname = "bpf_wq"


class HidRdescDescriptor(ctypes.Structure):
    cname = "hid_rdesc_descriptor"


class TestAsyncCb(ctypes.Structure):
    cname = "test_async_cb"
    _fields_overrides_ = {
        "cb": ctypes.c_void_p,  # just too complex to handle function pointers
    }


@dataclass
class OutputReport:
    time: int
    data: Tuple[bytes]


class ReportType(IntEnum):
    INPUT = 0
    OUTPUT = 1
    FEATURE = 2


class RequestType(IntEnum):
    GET_REPORT = 0x01
    GET_IDLE = 0x02
    GET_PROTOCOL = 0x03
    SET_REPORT = 0x09
    SET_IDLE = 0x0A
    SET_PROTOCOL = 0x0B


@dataclass
class HidRawRequest:
    time: int
    req_data: Tuple[bytes]
    out_data: Tuple[bytes]
    report_type: ReportType
    request_type: RequestType


@dataclass
class InputReport:
    time: int
    data: Tuple[bytes]
    report_type: ReportType


@dataclass
class PrivateBpfNumIterator:
    start: int
    end: int
    current_value: Optional[int] = dataclasses.field(default=None)

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_value is None:
            self.current_value = self.start
        else:
            self.current_value += 1

        if self.current_value >= self.end:
            raise StopIteration

        return ctypes.c_int(self.current_value)


@dataclass
class PrivateTestData:
    bpf: "Bpf"
    current_ctx: HidBpfCtx = dataclasses.field(default_factory=HidBpfCtx)
    id: int = dataclasses.field(default_factory=lambda: random.randint(0, 0xFFFF))
    vendor: int = 0
    product: int = 0
    output_reports: list[OutputReport] = dataclasses.field(default_factory=list)
    hw_requests: list[HidRawRequest] = dataclasses.field(default_factory=list)
    input_reports: list[InputReport] = dataclasses.field(default_factory=list)
    maps_data: dict[int, dict[int, ...]] = dataclasses.field(default_factory=dict)
    queue_data: dict[int, collections.deque] = dataclasses.field(default_factory=dict)
    asyncs: dict[int, TestAsyncCb] = dataclasses.field(default_factory=dict)
    iterators: dict[int, PrivateBpfNumIterator] = dataclasses.field(
        default_factory=dict
    )


# see struct test_callbacks
class Callbacks(ctypes.Structure):
    cname = "test_callbacks"
    _fields_overrides_ = {
        "private_data": ctypes.py_object,
    }

    def __init__(self, private: PrivateTestData):
        hid = HidDevice(id=private.id)
        hid.vendor = private.vendor
        hid.product = private.product
        private.current_ctx.hid = ctypes.pointer(hid)

        super().__init__(private_data=ctypes.py_object(private))
        fun_type = ctypes.CFUNCTYPE(None)
        for field, argtype in self._fields_:
            if isinstance(argtype, type(fun_type)):
                fun = getattr(Callbacks, f"_{field}")
                setattr(self, field, argtype(fun))

    def _hid_bpf_allocate_context(callbacks_p, hid: int):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data
        assert pdata.id == hid

        callbacks.ctx = ctypes.pointer(pdata.current_ctx)

        return 0

    def _hid_bpf_release_context(callbacks_p, ctx):
        callbacks = callbacks_p.contents

        callbacks.ctx = None

    @classmethod
    def validate_ctx(cls, callbacks_p, ctx_p):
        callbacks = callbacks_p.contents
        hid = ctx_p.contents.hid.contents
        return hid.id == callbacks.private_data.id

    def _hid_bpf_hw_output_report(callbacks_p, ctx_p, data_p, size):
        if not Callbacks.validate_ctx(callbacks_p, ctx_p):
            return -errno.EINVAL
        DataArray = ctypes.c_uint8 * size
        c_data = DataArray()
        p2 = ctypes.byref(c_data)
        ctypes.memmove(p2, data_p, size)
        data = bytes(c_data)

        callbacks = callbacks_p.contents
        callbacks.private_data.output_reports.append(OutputReport(callbacks.time, data))
        return size

    def _hid_bpf_hw_request(callbacks_p, ctx_p, data_p, size, _type, reqtype):
        if not Callbacks.validate_ctx(callbacks_p, ctx_p):
            return -errno.EINVAL
        DataArray = ctypes.c_uint8 * size
        c_data = DataArray()
        p2 = ctypes.byref(c_data)
        ctypes.memmove(p2, data_p, size)
        req_data = bytes(c_data)
        out_data = bytes(c_data)

        callbacks = callbacks_p.contents
        callbacks.private_data.hw_requests.append(
            HidRawRequest(
                callbacks.time,
                req_data,
                out_data,
                ReportType(_type),
                RequestType(reqtype),
            )
        )
        return size

    def _bpf_map_lookup_elem(callbacks_p, map_p, key_p):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]

        # ensure our type is known
        pdata.bpf.btf.build_struct(map.ctype)

        # allocate / get memory
        map_data_dict = pdata.maps_data.setdefault(map_p, {})
        key = ctypes.c_int.from_address(key_p).value
        data = map_data_dict.setdefault(key, map.ctype())

        # store the returned value in callbacks
        callbacks.helpers_retval = ctypes.cast(ctypes.byref(data), ctypes.c_void_p)

        return 0

    def _bpf_map_pop_elem(callbacks_p, map_p, data_p):
        callbacks = callbacks_p.contents
        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        queue = pdata.queue_data.get(map_p)
        if not queue:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]
        value = queue.popleft()
        ctypes.memmove(data_p, ctypes.byref(value), ctypes.sizeof(map.ctype))

        return 0

    def _bpf_map_push_elem(callbacks_p, map_p, data_p, flags):
        callbacks = callbacks_p.contents
        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]
        queue = pdata.queue_data.setdefault(map_p, collections.deque())

        value = map.ctype()
        ctypes.memmove(ctypes.byref(value), data_p, ctypes.sizeof(value))
        queue.append(value)

        return 0

    @classmethod
    def _async_init(cls, callbacks_p, async_p, map_p, clock, async_type):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]
        map_data_dict = pdata.maps_data[map_p]

        async_field = [
            name for name, value in map.ctype._fields_ if value == async_type
        ]

        # this ensures the provided map_p is correct
        assert len(async_field) == 1

        async_field = async_field.pop()

        for key, value in map_data_dict.items():
            t = ctypes.byref(getattr(value, async_field))
            t_p = ctypes.cast(t, ctypes.c_void_p).value
            if async_p == t_p:
                pdata.asyncs[async_p] = TestAsyncCb(
                    map_p,
                    key,
                    ctypes.cast(ctypes.byref(value), ctypes.c_void_p),
                    None,
                )
                return 0

        return -errno.EINVAL

    def _async_set_callback(callbacks_p, async_p, cb):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if async_p not in pdata.asyncs:
            return -errno.EINVAL

        _async = pdata.asyncs[async_p]
        _async.cb = ctypes.c_void_p(cb)

        return 0

    def _async_start(callbacks_p, async_p, delay, flags):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if async_p not in pdata.asyncs:
            return -errno.EINVAL

        _async = pdata.asyncs[async_p]

        assert _async.cb is not None

        # store the returned value in callbacks
        callbacks.helpers_retval = ctypes.cast(ctypes.byref(_async), ctypes.c_void_p)

        return 0

    def _bpf_timer_init(callbacks_p, timer_p, map_p, clock):
        return Callbacks._async_init(callbacks_p, timer_p, map_p, clock, BpfTimer)

    def _bpf_wq_init(callbacks_p, wq_p, map_p, clock):
        return Callbacks._async_init(callbacks_p, wq_p, map_p, clock, BpfWq)

    def _bpf_iter_num_new(callbacks_p, iter_p, start, end):
        """
        Initialize a number iterator.
        iter_p points to a bpf_iter_num structure.
        Returns 0 on success.
        """
        callbacks = callbacks_p.contents
        pdata = callbacks.private_data

        pdata.iterators[iter_p] = PrivateBpfNumIterator(start, end)
        return 0

    def _bpf_iter_num_next(callbacks_p, iter_p):
        """
        Get next value from number iterator.
        Returns pointer to current value, or NULL if done.
        """
        callbacks = callbacks_p.contents
        pdata = callbacks.private_data

        if iter_p not in pdata.iterators:
            return None

        iter_state = pdata.iterators[iter_p]

        # Get next value from generator
        current = next(iter_state, None)
        if current is None:
            return None

        callbacks.helpers_retval = ctypes.cast(
            ctypes.byref(current), ctypes.c_void_p
        ).value
        return callbacks.helpers_retval

    def _bpf_iter_num_destroy(callbacks_p, iter_p):
        """
        Clean up number iterator.
        Returns 0 on success.
        """
        callbacks = callbacks_p.contents
        pdata = callbacks.private_data

        if iter_p in pdata.iterators:
            del pdata.iterators[iter_p]

        return 0

    def _hid_bpf_input_report(callbacks_p, ctx_p, _type, data_p, size):
        if not Callbacks.validate_ctx(callbacks_p, ctx_p):
            return -errno.EINVAL
        DataArray = ctypes.c_uint8 * size
        c_data = DataArray()
        p2 = ctypes.byref(c_data)
        ctypes.memmove(p2, data_p, size)
        data = bytes(c_data)

        callbacks = callbacks_p.contents
        callbacks.private_data.input_reports.append(
            InputReport(callbacks.time, data, ReportType(_type))
        )
        return 0


@dataclass
class Api:
    """
    Wrapper to make automatically loading functions from a .so file simpler.
    """

    name: str
    args: Tuple[Type[ctypes._SimpleCData | ctypes._Pointer], ...]
    return_type: Optional[Type[ctypes._SimpleCData | ctypes._Pointer]]
    optional: bool

    @property
    def basename(self) -> str:
        return f"_{self.name}"


class Bpf:
    _api_prototypes: list[Api] = [
        Api(
            name="probe",
            args=(ctypes.POINTER(HidProbeArgs),),
            return_type=c_int,
            optional=True,
        ),
        Api(
            name="set_callbacks",
            args=(ctypes.POINTER(Callbacks),),
            return_type=None,
            optional=False,
        ),
    ]

    def __init__(self, name: str):
        lib_name = f"libtest-{name}"

        # Our test setup guarantees this works, running things manually is
        # a bit more complicated.
        ld_path = os.environ.get("LD_LIBRARY_PATH")
        assert ld_path is not None, (
            "Expected LD_LIBRARY_PATH to be set up for the tests"
        )

        sofile = Path(ld_path) / f"{lib_name}.so"
        if not sofile.exists():
            pytest.skip(f"Unable to locate {sofile}, assuming this BPF wasn't built")

        btf = _load_btf(name)
        for c in [
            HidProbeArgs,
            HidDevice,
            HidBpfCtx,
            BpfTimer,
            BpfWq,
            TestAsyncCb,
            Callbacks,
        ]:
            btf.build_struct(c)
            assert hasattr(c, "_fields_")

        # HidRdescDescriptor is optional - only present in BPFs that use
        # HID_REPORT_DESCRIPTOR
        btf.build_struct(HidRdescDescriptor)

        jsonfile = Path(ld_path) / f"{lib_name}.json"
        if not jsonfile.exists():
            pytest.skip(f"Unable to locate {jsonfile}, assuming this BPF wasn't built")

        # Load the libtest-$BPF.so file first.o, map probe and set_callbacks which
        # have a fixed name.
        #
        # Then try to find the corresponding libtest-$BPF.json file that meson
        # should have generated.
        # Because our actual entry points have custom names we check the json for the
        # right section and then map those we want into fixed-name wrappers, i.e.
        # SEC(HID_BPF_RDESC_FIXUP) becomes self._hid_bpf_rdesc_fixup() which points
        # to the right ctypes function.
        try:
            lib = ctypes.CDLL(sofile.name, use_errno=True)
            assert lib is not None
        except OSError as e:
            pytest.exit(
                f"Error loading the library: {e}. Maybe export LD_LIBRARY_PATH=builddir/test"
            )
        for api in self._api_prototypes:
            if api.optional and not hasattr(lib, api.name):
                continue
            func = getattr(lib, api.name)
            func.argtypes = api.args
            func.restype = api.return_type
            setattr(lib, api.basename, func)

        maps = {
            ctypes.cast(getattr(lib, m.name), ctypes.c_void_p).value: m
            for m in btf.maps
        }

        try:
            with open(jsonfile) as f:
                js = json.load(f)[0]
            for program in js["programs"]:

                def register_fun(generic_name):
                    func = getattr(lib, program["name"])
                    func.argtypes = (ctypes.POINTER(HidBpfCtx),)
                    func.restype = c_int
                    setattr(lib, generic_name, func)

                if program["section"].endswith("/hid_bpf_rdesc_fixup") or program[
                    "section"
                ].endswith("/hid_rdesc_fixup"):
                    register_fun("_hid_bpf_rdesc_fixup")
                elif program["section"].endswith("/hid_bpf_device_event") or program[
                    "section"
                ].endswith("/hid_device_event"):
                    register_fun("_hid_bpf_device_event")
        except OSError as e:
            pytest.exit(
                f"Error loading the JSON file: {e}. Unexpected LD_LIBRARY_PATH?"
            )

        self.lib = lib
        self._callbacks = None
        self.maps = maps
        self.btf = btf

    def close(self):
        """Close the shared library handle."""
        import _ctypes

        if self.lib is not None:
            _ctypes.dlclose(self.lib._handle)
            self.lib = None

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def get_global_u32(self, name: str) -> int:
        """
        Get a global u32 variable from the loaded BPF program.

        Example:
            value = bpf.get_global_u32("test_extract_8bit_result")
        """
        try:
            var = ctypes.c_uint32.in_dll(self.lib, name)
            return var.value
        except (ValueError, AttributeError) as e:
            raise KeyError(f"Global variable '{name}' not found in BPF program") from e

    def set_global_u32(self, name: str, value: int):
        """
        Set a global u32 variable in the loaded BPF program.

        Example:
            bpf.set_global_u32("test_variable", 42)
        """
        try:
            var = ctypes.c_uint32.in_dll(self.lib, name)
            var.value = value
        except (ValueError, AttributeError) as e:
            raise KeyError(f"Global variable '{name}' not found in BPF program") from e

    def clear_report_descriptor(self) -> bool:
        """
        Clear the HID_REPORT_DESCRIPTOR global variable by memsetting it to 0.

        Returns True if clearing was performed, False if HID_REPORT_DESCRIPTOR doesn't exist.
        """
        # Check if HID_REPORT_DESCRIPTOR exists in the loaded library
        try:
            var_struct = HidRdescDescriptor.in_dll(self.lib, "HID_REPORT_DESCRIPTOR")
        except (ValueError, AttributeError):
            return False

        try:
            struct_size = ctypes.sizeof(HidRdescDescriptor)

            ctypes.memset(ctypes.addressof(var_struct), 0, struct_size)

            logger.warning(f"Cleared {struct_size} bytes of HID_REPORT_DESCRIPTOR")
            return True

        except Exception as e:
            logger.error(f"Failed to clear HID_REPORT_DESCRIPTOR: {e}")
            return False

    def inject_report_descriptor(self, rdesc_bytes: bytes) -> bool:
        """
        Inject parsed HID report descriptor into HID_REPORT_DESCRIPTOR global variable.

        This method:
        1. Checks if HID_REPORT_DESCRIPTOR exists in the loaded BPF program
        2. If it exists, uses hid-rdesc-to-c-struct to convert rdesc_bytes to C-struct
        3. Injects the C-struct data into the HID_REPORT_DESCRIPTOR global variable

        Returns True if injection was performed, False if HID_REPORT_DESCRIPTOR doesn't exist.
        """
        try:
            var_struct = HidRdescDescriptor.in_dll(self.lib, "HID_REPORT_DESCRIPTOR")
        except (ValueError, AttributeError):
            return False

        # Convert rdesc_bytes to C-struct using the tool
        try:
            result = subprocess.run(
                ["hid-rdesc-to-c-struct"],  # base64 is the default format
                input=rdesc_bytes,
                capture_output=True,
                check=True,
            )
            base64_output = result.stdout.decode("utf-8").strip()

            c_struct_bytes = base64.b64decode(base64_output)

            # Get the actual size of the HID_REPORT_DESCRIPTOR struct.
            # We already checked that it was matching the BTF so we know
            # it is the actual binary size.
            struct_size = ctypes.sizeof(HidRdescDescriptor)

            if len(c_struct_bytes) > struct_size:
                logger.error(
                    f"Parsed descriptor size ({len(c_struct_bytes)} bytes) exceeds "
                    f"HID_REPORT_DESCRIPTOR size ({struct_size} bytes)"
                )
                return False

            ctypes.memmove(
                ctypes.addressof(var_struct), c_struct_bytes, len(c_struct_bytes)
            )

            logger.debug(
                f"Injected {len(c_struct_bytes)} bytes into HID_REPORT_DESCRIPTOR ({struct_size} bytes total)"
            )
            return True

        except OSError as e:
            logger.warning(
                f"Failed to run hid-rdesc-to-c-struct: {e}, skipping HID_REPORT_DESCRIPTOR injection"
            )
            return False
        except subprocess.CalledProcessError as e:
            logger.error(
                f"Failed to convert report descriptor: {e.stderr.decode('utf-8')}"
            )
            return False
        except Exception as e:
            logger.error(f"Failed to inject HID_REPORT_DESCRIPTOR: {e}")
            return False

    def set_callbacks(self, callbacks: Callbacks):
        """
        Set the callbacks to use for the various hid_bpf_* functions that may be
        used by a BPF program. These need to have a matching implementation in
        test-wrapper.c

        For most tests this isn't needed and you can pass the rdesc/report bytes
        directly to hid_bpf_rdesc_fixup() or hid_bpf_device_event().
        """
        self.lib._set_callbacks(callbacks)

    def probe(
        self,
        probe_args: HidProbeArgs,
        private_data: PrivateTestData | None = None,
    ) -> HidProbeArgs:
        """Call the BPF program's probe() function"""
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        callbacks = Callbacks(private_data)
        self.set_callbacks(callbacks)

        # We copy so our caller's probe args are separate from
        # the ones we return after the BPF program modifies them.
        pa = HidProbeArgs()
        p1 = ctypes.byref(probe_args)
        p2 = ctypes.byref(pa)
        ctypes.memmove(p2, p1, ctypes.sizeof(HidProbeArgs))
        pa.hid = callbacks.private_data.id

        # Extract rdesc bytes from probe_args and inject parsed descriptor if needed
        rdesc_size = probe_args.rdesc_size
        if rdesc_size > 0:
            rdesc_bytes = bytes(probe_args.rdesc[:rdesc_size])
            logger.info(f"Injecting {rdesc_size} byte report descriptor before probe")
            injected = self.inject_report_descriptor(rdesc_bytes)
            logger.info(f"Injection {'succeeded' if injected else 'skipped/failed'}")
            if not injected:
                self.clear_report_descriptor()
        else:
            self.clear_report_descriptor()

        rc = self.lib._probe(ctypes.byref(pa))
        if rc != 0:
            raise OSError(rc)
        return pa

    def hid_bpf_device_event(
        self,
        report: bytes | str | None = None,
        private_data: PrivateTestData | None = None,
    ) -> None | bytes:
        """
        Call the BPF program's hid_bpf_device_event function.

        If a report is given, it returns the (possibly modified) report.
        Otherwise it returns None.

        If the report is given with a string representation, it returns
        a string representation as well instead of a byte array.
        """
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        ctx = private_data.current_ctx

        input_is_binary = isinstance(report, bytes) or isinstance(report, bytearray)

        if report is not None:
            binary_report = (
                report
                if input_is_binary
                else binascii.unhexlify(report.replace(" ", ""))
            )
            allocated_size = int(len(binary_report) / 64 + 1) * 64
            data = (ctypes.c_uint8 * allocated_size)(*binary_report)
            callbacks = Callbacks(private_data)
            callbacks.hid_bpf_data = data
            callbacks.hid_bpf_data_sz = allocated_size
            ctx.allocated_size = allocated_size
            ctx.size = len(binary_report)
            self.set_callbacks(callbacks)
        else:
            data = None

        rc = self.lib._hid_bpf_device_event(ctypes.byref(ctx))
        if rc < 0:
            raise OSError(-rc)

        if rc > 0:
            ctx.retval = rc

        if report is None:
            return None
        assert data is not None

        if input_is_binary:
            return bytes(data[: ctx.retval])

        return binascii.hexlify(bytes(data[: ctx.retval]), " ").decode()

    def hid_bpf_rdesc_fixup(
        self,
        rdesc: bytes | str | None = None,
        private_data: PrivateTestData | None = None,
    ) -> None | bytes:
        """
        Call the BPF program's hid_bpf_rdesc_fixup function.

        If an rdesc is given, it returns the (possibly modified) rdesc.
        Otherwise it returns None.

        If the rdesc is given with a string representation, it returns
        a string representation as well instead of a byte array.
        """
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        ctx = private_data.current_ctx

        input_is_binary = isinstance(rdesc, bytes) or isinstance(rdesc, bytearray)

        if rdesc is not None:
            binary_rdesc = (
                rdesc if input_is_binary else binascii.unhexlify(rdesc.replace(" ", ""))
            )
            allocated_size = 4096
            data = (ctypes.c_uint8 * allocated_size)(*binary_rdesc)
            callbacks = Callbacks(private_data)
            callbacks.hid_bpf_data = data
            callbacks.hid_bpf_data_sz = allocated_size
            ctx.allocated_size = allocated_size
            ctx.size = len(binary_rdesc)
            self.set_callbacks(callbacks)
        else:
            data = None

        rc = self.lib._hid_bpf_rdesc_fixup(ctypes.byref(ctx))
        if rc < 0:
            raise OSError(rc)

        if rc > 0:
            ctx.retval = rc

        if rdesc is None:
            return None
        assert data is not None

        if input_is_binary:
            return bytes(data[: ctx.retval])

        return binascii.hexlify(bytes(data[: ctx.retval]), " ").decode()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
#
# Helper script to convert the jinja file into our hut.rs file
#
# Use as:
#     $ python3 ./tools/generate-hid-report-helpers.py \
#                        --datafile data/HidUsageTables.json \
#                        ./tools/hid_report_helpers.h.jinja > src/bpf/hid_report_helpers.h
#
# Or (for hid_usages.h):
#     $ python3 ./tools/generate-hid-report-helpers.py \
#                        --datafile data/HidUsageTables.json \
#                        ./tools/hid_usages.h.jinja > src/bpf/hid_usages.h

from pathlib import Path
from dataclasses import dataclass
import argparse
import enum
import os
import re
import sys
import functools
import json
import jinja2
import jinja2.environment


def sanitize(s):
    s = re.sub(r"[-‐\t: \\|/.\n,()]", "", s)
    s = re.sub(r"[+]", "Plus", s)
    s = re.sub(r"[-]", "Minus", s)
    numbers = {
        "1": "One",
        "2": "Two",
        "3": "Three",
        "4": "Four",
        "5": "Five",
        "6": "Six",
        "7": "Seven",
        "8": "Eight",
        "9": "Nine",
        "0": "Zero",
    }
    s = functools.reduce(
        lambda s, kv: re.sub(rf"^({kv[0]})", kv[1], s), numbers.items(), s
    )
    return s


def sanitize_keyboard(s):
    map = {
        "!": "ExclamationMark",
        "@": "At",
        "/": "Slash",
        "#": "Hash",
        "$": "Dollar",
        "%": "Percent",
        "^": "Caret",
        "&": "Ampersand",
        "*": "Star",
        "[": "OpenBracket",
        "]": "CloseBracket",
        "{": "OpenBrace",
        "}": "CloseBrace",
        "(": "OpenParenthesis",
        ")": "CloseParenthesis",
        "'": "SingleQuote",
        '"': "DoubleQuote",
        "<": "LessThan",
        ">": "GreaterThan",
        " and ": "And",
        " ": "",
        ",": "Comma",
        "-": "Minus",
        "+": "Plus",
        "|": "Pipe",
        "?": "QuestionMark",
        "=": "Equals",
        "\\": "Backslash",
        "~": "Tilde",
        ";": "Semicolon",
        ":": "Colon",
        ".": "Period",
    }
    s = functools.reduce(lambda s, kv: s.replace(kv[0], kv[1]), map.items(), s)

    numbers = {
        "1": "One",
        "2": "Two",
        "3": "Three",
        "4": "Four",
        "5": "Five",
        "6": "Six",
        "7": "Seven",
        "8": "Eight",
        "9": "Nine",
        "0": "Zero",
    }
    s = functools.reduce(
        lambda s, kv: re.sub(rf"^({kv[0]})", kv[1], s), numbers.items(), s
    )
    return s


class UsageType(enum.StrEnum):
    LINEAR_CONTROL = "LC"
    ON_OFF_CONTROL = "OOC"
    MOMENTARY_CONTROL = "MC"
    ONE_SHOT_CONTROL = "OSC"
    RETRIGGER_CONTROL = "RTC"
    SELECTOR = "Sel"
    STATIC_VALUE = "SV"
    STATIC_FLAG = "SF"
    DYNAMIC_VALUE = "DV"
    DYNAMIC_FLAG = "DF"
    NAMED_ARRAY = "NAry"
    APPLICATION_COLLECTION = "CA"
    LOGICAL_COLLECTION = "CL"
    PHYSICAL_COLLECTION = "CP"
    USAGE_SWITCH = "US"
    USAGE_MODIFIER = "UM"
    BUFFERED_BYTES = "BufferedBytes"


class UsagePageType(enum.StrEnum):
    DEFINED = "Defined"
    GENERATED = "Generated"


@dataclass
class UsageId:
    printable: str  # The string as it was in the data file
    name: str  # Name of the usage sanitized to be a valid identifier
    value: int
    usage_types: list[UsageType]


@dataclass
class UsagePage:
    printable: str
    name: str
    acronym: str
    value: int
    usages: list[UsageId]
    usage_page_type: UsagePageType
    name_prefix: str | None


def parse_usage_id(u) -> UsageId:
    # Some of the strings have \_ in them for whatever reason...
    printable = u["Name"].replace("\\", "")
    name = sanitize(printable)
    value = int(u["Id"])
    types = [UsageType(k) for k in u["Kinds"]]
    usage = UsageId(printable=printable, name=name, value=value, usage_types=types)
    return usage


def parse_usage_page(up) -> UsagePage:
    printable = up["Name"]
    name = sanitize(printable)
    acronym = "".join([x for x in name if x.isupper()])
    if len(acronym) == 1:
        acronym = name[:3]
    value = int(up["Id"])
    usages = sorted([parse_usage_id(u) for u in up["UsageIds"]], key=lambda u: u.value)
    uptype = UsagePageType(up["Kind"])
    name_prefix = None
    if uptype == UsagePageType.GENERATED:
        name_prefix = up["UsageIdGenerator"]["NamePrefix"]
        if name_prefix.lower() == "enum":
            name_prefix += "erate"
    usage_page = UsagePage(
        printable=printable,
        name=name,
        acronym=acronym,
        value=value,
        usages=usages,
        usage_page_type=uptype,
        name_prefix=name_prefix,
    )
    return usage_page


def parse_data_file(datadir):
    js = json.load(open(datadir, "r"))
    usage_pages = [parse_usage_page(up) for up in js["UsagePages"]]

    # For some reason the Unicode usage page isn't in the JSON but
    # it is described on page 213 of the HUT 1.5 document. Let's insert
    # it manually.
    unicode_page = UsagePage(
        printable="Unicode",
        name="Unicode",
        acronym="UC",
        value=0x10,
        usages=[],
        usage_page_type=UsagePageType.GENERATED,
        name_prefix="codepoint",
    )
    usage_pages.append(unicode_page)

    return sorted(usage_pages, key=lambda up: up.value)


def generate_source(
    usage_pages: list[UsagePage], template: str
) -> jinja2.environment.TemplateStream:
    data = {}
    data["usage_pages"] = usage_pages

    loader: jinja2.BaseLoader
    if template == "-":
        loader = jinja2.FunctionLoader(lambda _: sys.stdin.read())
        filename = "<stdin>"
    else:
        path = Path(template)
        assert path.exists(), f"Failed to find template {path}"
        filename = path.name
        loader = jinja2.FileSystemLoader(os.fspath(path.parent))

    env = jinja2.Environment(
        loader=loader,
        trim_blocks=True,
        lstrip_blocks=True,
    )
    jtemplate = env.get_template(filename)
    return jtemplate.stream(data)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", default="-")
    parser.add_argument(
        "--datafile",
        type=Path,
        default=Path("data"),
        help="Path to the HUT json",
    )
    parser.add_argument("template", type=Path, help="The jinja template file")
    args = parser.parse_args()
    assert args.datafile.exists()
    assert args.template.exists()

    hut = parse_data_file(args.datafile)

    stream = generate_source(usage_pages=hut, template=args.template)
    file = sys.stdout if args.output == "-" else open(args.output, "w")
    stream.dump(file)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2024 Red Hat

"""
Shared pytest fixtures for BPF testing.
"""

from . import Bpf
import pytest


@pytest.fixture
def bpf(source: str):
    """
    A fixture that allows parametrizing over a number of sources. Use with e.g.::

        @pytest.mark.parametrize("source", ["0010-FR-TEC__Raptor-Mach-2"])
        def test_something(bpf):
            pass
    """
    assert source is not None
    with Bpf(source) as bpf:
        yield bpf



#!/usr/bin/env python3
#
# SPDX-License-Identifier: GPL-2.0-only
#
# Script to help create a release tarball for udev-hid-bpf.
# Copies all items in INSTALL_TREE to the given target directory
# and generates the install.sh/uninstall.sh files based on
# the files in that tree.
#
# INSTALL_TREE must be a clean install tree with nothing but the udev-hid-bpf
# files.

from dataclasses import dataclass
from pathlib import Path

import argparse
import jinja2
import jinja2.environment
import os
import shutil
import logging


logger = logging.getLogger("")
logging.basicConfig(level=logging.DEBUG, format="%(levelname)s: %(message)s")
logger.setLevel(logging.INFO)

parser = argparse.ArgumentParser(
    description="Creates a release tree from an installed udev-hid-bpf directory"
)
parser.add_argument("--verbose", action="store_true", default=False)
parser.add_argument(
    "--templates",
    action="append",
    type=Path,
    help="A jinja template to parse and place in the top-level directory of the TARGET_DIR",
)
parser.add_argument(
    "source_dir",
    metavar="INSTALL_TREE",
    type=Path,
    help="A meson install-ed tree of udev-hid-bpf",
)
parser.add_argument(
    "target_dir",
    metavar="TARGET_DIR",
    type=Path,
    help="Target directory to place all files",
)
args = parser.parse_args()

if args.verbose:
    logger.setLevel(logging.DEBUG)


@dataclass
class InstallFile:
    path: Path

    @property
    def dir(self) -> Path:
        return self.path.parent

    @property
    def filename(self) -> str:
        return self.path.name


bpfs = []
udev_rules = []
hwdb_files = []
# First copy our installed tree over
for file in args.source_dir.glob("**/*"):
    if file.is_dir():
        continue

    path = file.relative_to(args.source_dir)
    target = args.target_dir / path
    target.parent.mkdir(exist_ok=True, parents=True)
    logger.debug(f"{target}")
    shutil.copyfile(file, target)
    shutil.copymode(file, target)

    ifile = InstallFile(target.relative_to(args.target_dir))

    if target.name.endswith(".bpf.o"):
        bpfs.append(ifile)
    elif target.name.endswith(".rules"):
        udev_rules.append(ifile)
    elif target.name.endswith(".hwdb"):
        hwdb_files.append(ifile)


data = {
    "bpfs": bpfs,
    "udev_rules": udev_rules,
    "hwdb_files": hwdb_files,
    "pipeline_url": os.getenv("CI_PIPELINE_URL", "local invocation"),
}

for template in args.templates:
    template_dir = template.absolute().parent
    loader = jinja2.FileSystemLoader(template_dir)
    env = jinja2.Environment(
        loader=loader,
        trim_blocks=True,
        lstrip_blocks=True,
    )
    jtemplate = env.get_template(template.name)
    stream = jtemplate.stream(data)
    target = args.target_dir / template.name.rstrip(".jinja")
    logger.debug(target)
    with open(target, "w") as fd:
        stream.dump(fd)
    shutil.copymode(template, target)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2025 Red Hat

"""
Tests for new BPF helper functionality:
- HID report descriptor injection
- extract_bits optimizations (8/16/32-bit aligned fast paths)
- Iterator macros (hid_bpf_for_each_*)
"""

from . import HidProbeArgs
from dataclasses import dataclass
from enum import IntEnum
import pytest


class DeviceType(IntEnum):
    """HID device types."""

    MOUSE = 0x02
    KEYBOARD = 0x06


class ReportItemType(IntEnum):
    """HID report item types."""

    INPUT = 0x81
    OUTPUT = 0x91
    FEATURE = 0xB1


@dataclass(frozen=True)
class DeviceConfig:
    """Configuration for a specific HID device type."""

    usage: int
    usage_page: int
    usage_min: int
    usage_max: int
    log_min: int
    log_max: int


def create_hid_rdesc(
    device_type: DeviceType = DeviceType.MOUSE,
    report_type: ReportItemType = ReportItemType.INPUT,
    nested: bool = True,
) -> bytearray:
    """
    Create a HID report descriptor.

    Args:
        device_type: DeviceType enum value
        report_type: ReportItemType enum value
        nested: Whether to include nested Physical collection (for Mouse)

    Returns:
        bytearray containing the HID report descriptor
    """
    # Device-specific configurations
    devices = {
        DeviceType.MOUSE: DeviceConfig(
            usage=0x02,  # Mouse
            usage_page=0x09,  # Button
            usage_min=0x01,  # Button 1
            usage_max=0x03,  # Button 3
            log_min=0x00,
            log_max=0x01,
        ),
        DeviceType.KEYBOARD: DeviceConfig(
            usage=0x06,  # Keyboard
            usage_page=0x07,  # Keyboard
            usage_min=0x00,
            usage_max=0xFF,
            log_min=0x00,
            log_max=0xFF,
        ),
    }

    device = devices[device_type]
    report_item = report_type

    rdesc = bytearray(
        [
            0x05,
            0x01,  # Usage Page (Generic Desktop)
            0x09,
            device.usage,  # Usage (Mouse/Keyboard)
            0xA1,
            0x01,  # Collection (Application)
        ]
    )

    # Add nested Physical collection for Mouse if requested
    if nested and device_type == DeviceType.MOUSE:
        rdesc.extend(
            [
                0x09,
                0x01,  # Usage (Pointer)
                0xA1,
                0x00,  # Collection (Physical)
            ]
        )

    # Add report ID for feature/output reports
    if report_type in [ReportItemType.FEATURE, ReportItemType.OUTPUT]:
        rdesc.extend([0x85, 0x01])  # Report ID (1)

    # Add report fields
    rdesc.extend(
        [
            0x05,
            device.usage_page,  # Usage Page
            0x19,
            device.usage_min,  # Usage Minimum
            0x29,
            device.usage_max,  # Usage Maximum
            0x15,
            device.log_min,  # Logical Minimum
            0x25,
            device.log_max,  # Logical Maximum
            0x75,
            0x08,  # Report Size (8)
            0x95,
            0x01,  # Report Count (1)
            report_item,
            0x02,  # Input/Output/Feature (Data,Var,Abs)
        ]
    )

    # Close nested Physical collection if opened
    if nested and device_type == DeviceType.MOUSE:
        rdesc.extend([0xC0])  # End Collection

    # Close Application collection
    rdesc.extend([0xC0])  # End Collection

    return rdesc


@pytest.mark.parametrize("source", ["0010-test-helpers"])
class TestExtractBits:
    """Test extract_bits function with various alignments and sizes."""

    @pytest.mark.parametrize(
        "bits_start,bits_end,byte_data,expected",
        [
            pytest.param(8, 16, [(1, 0x42)], 0x42, id="8bit_aligned"),
            pytest.param(16, 32, [(2, 0x34), (3, 0x12)], 0x1234, id="16bit_aligned"),
            pytest.param(
                32,
                64,
                [(4, 0x78), (5, 0x56), (6, 0x34), (7, 0x12)],
                0x12345678,
                id="32bit_aligned",
            ),
            pytest.param(4, 16, [(0, 0xC0), (1, 0xAB)], 0xABC, id="unaligned_12bit"),
            pytest.param(8, 16, [], 0, id="zero_8bit"),
            pytest.param(16, 32, [], 0, id="zero_16bit"),
            pytest.param(32, 64, [], 0, id="zero_32bit"),
        ],
    )
    def test_extract_bits(self, bpf, bits_start, bits_end, byte_data, expected):
        """Test extract_bits with various bit ranges and alignments."""
        # Set BPF test parameters
        bpf.set_global_u32("test_bits_start", bits_start)
        bpf.set_global_u32("test_bits_end", bits_end)

        # Create report with specified byte values
        report = bytearray(64)
        for offset, value in byte_data:
            report[offset] = value

        # Process the report
        result = bpf.hid_bpf_device_event(report=bytes(report))

        # Event passes through unchanged
        assert result == bytes(report)

        # Check the extracted value
        test_result = bpf.get_global_u32("test_extract_result")
        assert test_result == expected, (
            f"Extract bits[{bits_start}:{bits_end}] expected 0x{expected:X}, "
            f"got 0x{test_result:X}"
        )

    @pytest.mark.parametrize(
        "bits_start,bits_end,expected",
        [
            pytest.param(8, 16, 0xFF, id="max_8bit"),
            pytest.param(16, 32, 0xFFFF, id="max_16bit"),
            pytest.param(32, 64, 0xFFFFFFFF, id="max_32bit"),
        ],
    )
    def test_extract_max_values(self, bpf, bits_start, bits_end, expected):
        """Test extract_bits with all bits set to verify masking."""
        # Set BPF test parameters
        bpf.set_global_u32("test_bits_start", bits_start)
        bpf.set_global_u32("test_bits_end", bits_end)

        # Fill report with 0xFF
        report = bytearray([0xFF] * 64)

        # Process the report
        result = bpf.hid_bpf_device_event(report=bytes(report))
        assert result == bytes(report)

        # Check the extracted value
        test_result = bpf.get_global_u32("test_extract_result")
        assert test_result == expected, (
            f"Extract bits[{bits_start}:{bits_end}] with max values: "
            f"expected 0x{expected:X}, got 0x{test_result:X}"
        )


@pytest.mark.parametrize("source", ["0010-test-helpers"])
class TestIteratorMacros:
    """Test iterator macros for HID report descriptor traversal."""

    def test_report_descriptor_injection(self, bpf):
        """
        Test that the HID_REPORT_DESCRIPTOR is properly injected
        and accessible in the probe function.
        """
        probe_args = HidProbeArgs()
        probe_args.rdesc_bytes = create_hid_rdesc(
            device_type=DeviceType.MOUSE, report_type=ReportItemType.INPUT, nested=True
        )

        # Call probe - it should parse the descriptor and iterate
        pa = bpf.probe(probe_args)

        # Should complete successfully
        assert pa.retval == 0

    @pytest.mark.parametrize(
        "device_type,report_type",
        [
            (DeviceType.KEYBOARD, ReportItemType.FEATURE),
            (DeviceType.MOUSE, ReportItemType.INPUT),
        ],
    )
    def test_iterate_reports(self, bpf, device_type, report_type):
        """Test hid_bpf_for_each_{feature,input}_report macros."""
        # Map report type to counter variable
        count_vars = {
            ReportItemType.FEATURE: "test_feature_report_count",
            ReportItemType.INPUT: "test_input_report_count",
        }

        probe_args = HidProbeArgs()
        probe_args.rdesc_bytes = create_hid_rdesc(
            device_type=device_type, report_type=report_type, nested=False
        )

        pa = bpf.probe(probe_args)
        assert pa.retval == 0

        # Check that reports were counted
        count = bpf.get_global_u32(count_vars[report_type])
        assert count >= 0

    def test_iterate_fields(self, bpf):
        """Test hid_bpf_for_each_field macro."""
        probe_args = HidProbeArgs()

        pa = bpf.probe(probe_args)
        assert pa.retval == 0

        # Check that fields were counted
        field_count = bpf.get_global_u32("test_field_count")
        assert field_count >= 0

    def test_iterate_collections(self, bpf):
        """Test hid_bpf_for_each_collection macro."""
        probe_args = HidProbeArgs()
        # Use Mouse with nested Physical collection to test collection iteration
        probe_args.rdesc_bytes = create_hid_rdesc(
            device_type=DeviceType.MOUSE, report_type=ReportItemType.INPUT, nested=True
        )

        pa = bpf.probe(probe_args)
        assert pa.retval == 0

        # Check that collections were counted
        collection_count = bpf.get_global_u32("test_collection_count")
        assert collection_count > 0

    def test_empty_descriptor(self, bpf):
        """Test iterators with empty report descriptor."""
        probe_args = HidProbeArgs()

        # Should handle empty descriptor gracefully
        pa = bpf.probe(probe_args)

        # May return error or success, but shouldn't crash
        assert pa is not None

    def test_bounds_checking(self, bpf):
        """Test that iterators respect HID_MAX_* bounds."""
        # Define the HID_MAX_* constants from hid_bpf_helpers.h
        HID_MAX_REPORTS = 16
        HID_MAX_FIELDS = 64
        HID_MAX_COLLECTIONS = 32

        # Generate a descriptor that exceeds these limits
        rdesc_data = bytearray()

        # Start main collection
        rdesc_data.extend(
            [
                0x05,
                0x01,  # Usage Page (Generic Desktop)
                0x09,
                0x06,  # Usage (Keyboard)
                0xA1,
                0x01,  # Collection (Application)
            ]
        )

        # Generate HID_MAX_REPORTS + 4 input reports (exceeds HID_MAX_REPORTS)
        for report_id in range(1, HID_MAX_REPORTS + 5):
            rdesc_data.extend(
                [
                    0x85,
                    report_id,  # Report ID
                ]
            )

            # Report 1 has HID_MAX_FIELDS + 6 fields (exceeds HID_MAX_FIELDS)
            if report_id == 1:
                # Add individual 1-bit fields
                for field_num in range(HID_MAX_FIELDS + 6):
                    rdesc_data.extend(
                        [
                            0x05,
                            0x09,  # Usage Page (Button)
                            0x09,
                            (field_num % 255) + 1,  # Usage (Button N)
                            0x15,
                            0x00,  # Logical Minimum (0)
                            0x25,
                            0x01,  # Logical Maximum (1)
                            0x75,
                            0x01,  # Report Size (1)
                            0x95,
                            0x01,  # Report Count (1)
                            0x81,
                            0x02,  # Input (Data,Var,Abs)
                        ]
                    )
                # Padding to byte boundary
                remaining_bits = (HID_MAX_FIELDS + 6) % 8
                if remaining_bits:
                    rdesc_data.extend(
                        [
                            0x75,
                            0x01,  # Report Size (1)
                            0x95,
                            8 - remaining_bits,  # Report Count
                            0x81,
                            0x01,  # Input (Const)
                        ]
                    )
            else:
                # Other reports have simple 1-byte fields
                rdesc_data.extend(
                    [
                        0x05,
                        0x09,  # Usage Page (Button)
                        0x09,
                        0x01,  # Usage (Button 1)
                        0x15,
                        0x00,  # Logical Minimum (0)
                        0x25,
                        0xFF,  # Logical Maximum (255)
                        0x75,
                        0x08,  # Report Size (8)
                        0x95,
                        0x01,  # Report Count (1)
                        0x81,
                        0x02,  # Input (Data,Var,Abs)
                    ]
                )

        # Add HID_MAX_COLLECTIONS + 3 nested collections (exceeds HID_MAX_COLLECTIONS)
        for coll_num in range(HID_MAX_COLLECTIONS + 3):
            rdesc_data.extend(
                [
                    0xA1,
                    0x00,  # Collection (Physical)
                ]
            )

        # Close all nested collections
        for coll_num in range(HID_MAX_COLLECTIONS + 3):
            rdesc_data.extend(
                [
                    0xC0,  # End Collection
                ]
            )

        # Close main collection
        rdesc_data.extend(
            [
                0xC0,  # End Collection
            ]
        )

        probe_args = HidProbeArgs()
        probe_args.rdesc_bytes = rdesc_data

        # Even with a larger descriptor, iterators should
        # stop at HID_MAX_REPORTS, HID_MAX_FIELDS, etc.
        pa = bpf.probe(probe_args)

        assert pa.retval == 0

        # Verify that counts are capped at their maximums
        # We generated HID_MAX_REPORTS + 4 reports, but should only see HID_MAX_REPORTS
        input_report_count = bpf.get_global_u32("test_input_report_count")
        assert input_report_count == HID_MAX_REPORTS, (
            f"Expected {HID_MAX_REPORTS} input reports, got {input_report_count}"
        )

        # We generated HID_MAX_FIELDS + 6 fields in report 1, but should only see HID_MAX_FIELDS
        # Note: test_field_count is the total across all input reports
        # Report 1 should be capped at 64 fields
        field_count = bpf.get_global_u32("test_field_count")
        assert field_count >= HID_MAX_FIELDS, (
            f"Expected at least {HID_MAX_FIELDS} fields, got {field_count}"
        )

        # We generated HID_MAX_COLLECTIONS + 3 nested collections, but each field
        # should be capped at HID_MAX_COLLECTIONS
        max_collections_per_field = bpf.get_global_u32("test_max_collections_per_field")
        assert max_collections_per_field <= HID_MAX_COLLECTIONS, (
            f"Expected max {HID_MAX_COLLECTIONS} collections per field, got {max_collections_per_field}"
        )

        # Feature reports should be 0 since we only generated input reports
        feature_report_count = bpf.get_global_u32("test_feature_report_count")
        assert feature_report_count == 0, (
            f"Expected 0 feature reports, got {feature_report_count}"
        )
